
from .grammar import EOF
from .engine import Ambiguity
from .eng_sr import ShiftReduceEngine

class StackNode:
    def __init__(self, state):
        self.s = state
        self.l = {}

    def links(self):
        return self.l.items()

    def __repr__(self):
        return "<Node {} {}>".format(self.s, self.l)

class ActionNode (list):
    def resolve(self, eng):
        r = []
        for p,nodes,rnode in self:
            lnode = nodes[0]
            argv = []
            for n in range(1,len(nodes)):
                v = nodes[n].l[nodes[n-1]]
                if isinstance(v, ActionNode):
                    argv.append((None,v.resolve(eng)))
                else:
                    argv.append((None,v))
            r.append(eng.action(p, argv))
        if len(r) == 1:
            r = r[0]
        else:
            r = eng.merge_values(r)
        rnode.l[lnode] = r
        return r

class GLREngine (ShiftReduceEngine):
    def run(self, st=0):
        self.snode = StackNode(self.TABLE.states[st])
        self.stacks = { 0: self.snode }
        self.results = []
        self.parse()
        if len(self.results) == 0:
            raise self.error(self.t)
        assert len(self.results) == 1
        return self.results[0]

    def collect_reduces(self, a, node, ll):
        p = self.TABLE.rules[a]
        oll = [([node], ll)]
        for i in range(p.nsyms):
            nll = []
            for nodes, ll in oll:
                for node,v in ll:
                    nll.append(([node]+nodes, node.links()))
            oll = nll
        return [(p, nodes) for nodes,ll in oll]

    def add_link(self, lnode, rnode, v, t, epsilon):
        newl = (lnode, v)
        rnode.l[lnode] = v
        c,a = self.match(t, rnode.s)
        if c == 'r':
            if not epsilon and self.TABLE.rules[a].nsyms == 0:
                return []
            return self.collect_reduces(a, rnode, [newl])
        elif c != 'c':
            return []
        r = []
        for cc,aa in a:
            if cc == 'r':
                if epsilon or self.TABLE.rules[aa].nsyms:
                    r += self.collect_reduces(aa, rnode, [newl])
        return r

    def do_shift(self, lnode, t, a):
        if a in self.stacks:
            rnode = self.stacks[a]
        else:
            rnode = StackNode(self.TABLE.states[a])
            self.stacks[a] = rnode
        rnode.l[lnode] = t

    def parse(self):
        while self.stacks:
#            print()
#            print(', '.join(str(v.s.n) for v in self.stacks.values()))
            t = self.src.get()
#            print(t)
            self.t = t
            reduces = []
            for n in self.stacks.values():
                c,a = self.match(t, n.s)
                if c == 'r':
                    reduces += self.collect_reduces(a, n, n.links())
                elif c == 'c':
                    for cc,aa in a:
                        if cc == 'r':
                            reduces += self.collect_reduces(aa, n, n.links())
            acts = set()
            while reduces:
                p, nodes = reduces.pop(0)
                lnode = nodes[0]
                g = lnode.s.action[p.lhs][1]
                if g in self.stacks:
                    rnode = self.stacks[g]
                    if lnode in rnode.l:
                        v = rnode.l[lnode]
                        assert isinstance(v, ActionNode)
                        v.append((p,nodes,rnode))
                    else:
                        v = ActionNode(((p,nodes,rnode),))
                        reduces += self.add_link(lnode, rnode, v, t, False)
                else:
                    rnode = StackNode(self.TABLE.states[g])
                    self.stacks[g] = rnode
                    v = ActionNode(((p,nodes,rnode),))
                    reduces += self.add_link(lnode, rnode, v, t, True)
                acts.add((lnode,rnode))
            for lnode,rnode in acts:
                v = rnode.l[lnode]
                if isinstance(v, ActionNode):
                    v.resolve(self)
            ostacks = self.stacks
            self.stacks = {}
            for n in ostacks.values():
                c,a = self.match(t, n.s)
                if c == 'a':
                    assert len(n.l) == 1
                    self.results.append(n.l[self.snode])
                elif c == 's':
                    self.do_shift(n, t, a)
                elif c == 'c':
                    for cc,aa in a:
                        if cc == 'a':
                            assert len(n.l) == 1
                            self.results.append(n.l[self.snode])
                        elif cc == 's':
                            self.do_shift(n, t, aa)
        if t.type != EOF:
            raise self.error(t)

    def merge_values(self, vlist):
        ## assumes that values are Tokens
        t = vlist[0]
        t.value = Ambiguity(v.value for v in vlist)
        t.src = [ t.value ]
        return t
