
from .grammar import NonTerminal, ANY, TYP, TEXT, SET, TSYM, NTSYM, Symbol, \
    ensure_tables

class Preparation:
    def __init__(self, grammar):
        self.rules = []
        self.ntsyms = set()
        self.grammar = grammar
        self.run()

    def new_nt(self):
        nt = NonTerminal()
        self.ntsyms.add(nt)
        nt.name = "_{}".format(len(self.ntsyms))
        nt.anon = True
        return nt

class ParserGenerator:

    def __init__(self, grammar, pname, *pbases):
        self.grammar = grammar
        self.pname = pname
        self.pbases = pbases
        self._parser = None
        if hasattr(grammar, 'category'):
            self.category = grammar.category

    def prepare(self):
        ensure_tables()
        for n,p in enumerate(self.grammar.PRODUCTIONS):
            p.n = n
            p.ensure(self.grammar)
        self.pgr = self.PREPARATION(self.grammar)

    def category(self, c):
        return None

    def table(self):
        tbl = {}
        ss = self.start_states()
        todo = set(ss)
        while todo:
            ts = []
            st = todo.pop()
            stbl = {}
            tbl[st] = stbl
            for s in self.closure({st}):
                ts.extend(self.transitions(s))
            for t,ns in ts:
                if t in stbl and stbl[t] is not ns:
                    self.conflict(st, t, stbl[t], ns)
                elif ns:
                    stbl[t] = ns
                else:
                    stbl[t] = None
                if ns and ns not in tbl:
                    todo.add(ns)
        return ss, tbl

    def start_states(self):
        return [self.start_state(s) for s in self.pgr.start]

    def state_items(self, st):
        return self.closure({st})

    def parser(self):
        if self._parser is None:
            self.prepare()
            attrs = { 'TABLE': self.table() }
            bases = (self.grammar,)+self.pbases+(self.ENGINE,)
            self._parser = type(self.pname, bases, attrs)
        return self._parser

class SetsMethod:
    def start_states(self):
        return [self.make_state([s]) for s in super().start_states()]

    def make_state(self, ss):
        return frozenset(ss)

    def transitions(self, st):
        nts = {}
        ts = {}
        cs = {}
        any = set()
        dft = set()
        for s in self.state_items(st):
            for t in super().transitions(s):
                if isinstance(t[0], ANY):
                    any.add(t[1])
                elif isinstance(t[0], TYP):
                    cs.setdefault(t[0].c, set()).add(t[1])
                elif isinstance(t[0], SET):  # move this to transitions
                    if t[0].inv:
                        for v in t[0].s:
                            ts.setdefault(v, set())
                        dft.add(t[1])
                    else:
                        for v in t[0].s:
                            ts.setdefault(v, set()).add(t[1])
                elif isinstance(t[0], NTSYM):
                    nts.setdefault(t[0].v, set()).add(t[1])
                else:
                    ts.setdefault(t[0].v, set()).add(t[1])
        if any or dft:
            yield (ANY(), self.make_state(any | dft))
        for c in cs:
            yield (TYP(c), self.make_state(cs[c] | any | dft))
        for s in nts:
            yield (NTSYM(s), self.make_state(nts[s]))
        for s in ts:
            c = self.category(s)
            if isinstance(s, Symbol):
                ss = TSYM(s)
            else:
                ss = TEXT(s)
            yield (ss, self.make_state(ts[s] | cs.get(c, set()) | any))

    def closure(self, ss):
        return ss

    def state_items(self, st):
        return super().closure(st)
