
from .grammar import EOF
from .engine import Token

class FileSource:
    "Source of character symbols from a file."

    def __init__(self, f, filename="<file>", encoding=None):
        if isinstance(f, str):
            self.fn = f
            self.f = open(f)
        else:
            self.f = f
            self.fn = filename
        self.line = 0
        self.eof = False
        self.encoding = encoding

    def filename(self):
        return self.fn

    def get(self):
        if self.eof: return EOF
        l = self.f.readline()
        if l:
            self.line += 1
            if self.encoding:
                l = l.decode(self.encoding)
            return list(l)
        self.eof = True
        return EOF

class FileTokenSource (FileSource):
    def get(self):
        l = super().get()
        if self.eof: return Token('', EOF)
        return [Token(c) for c in l]

class FileNodeSourceTrack:
    def __init__(self, f, b, e):
        self.f = f
        self.b = b
        self.e = e

    def __or__(self, other):
        assert self.f == other.f
        if isinstance(other, FileSourceTrack):
            return FileNodeSourceTrack(self.f, self.b, (other.l,other.c+1))
        return FileNodeSourceTrack(self.f, self.b, other.e)

    def __repr__(self):
        return "{}:{}:{}-{}:{}".format(self.f, self.b[0], self.b[1],
                                       self.e[0], self.e[1])

class FileSourceTrack:
    def __init__(self, f, l, c):
        self.f = f
        self.l = l
        self.c = c

    def __or__(self, other):
        assert self.f == other.f
        if isinstance(other, FileSourceTrack):
            return FileNodeSourceTrack(self.f, (self.l,self.c),
                                       (other.l,other.c+1))
        return FileNodeSourceTrack(self.f, (self.l,self.c), other.e)

    def __str__(self):
        return "{}:{}:{}".format(self.f, self.l, self.c)

class FileTraceSource (FileTokenSource):
    lastc = 0

    def get(self):
        l = super().get()
        if self.eof:
            l.src = FileSourceTrack(self.fn, self.line, self.lastc)
            return l
        for n,t in enumerate(l): t.src=FileSourceTrack(self.fn, self.line, n)
        self.lastc = len(l)
        return l
