
from .grammar import EOF
from .engine import Token

class StringSource:
    "Source of character symbols from a string.  Treats the string as a single line."

    def __init__(self, s, filename="<string>"):
        self.fn = filename
        self.s = s
        self.eof = False

    def filename(self):
        return self.fn

    def get(self):
        if self.eof: return EOF
        self.eof = True
        return list(self.s)

class StringTokenSource (StringSource):
    def get(self):
        if self.eof: return Token('', EOF)
        self.eof = True
        return [Token(c) for c in self.s]

class StringNodeSourceTrack:
    def __init__(self, s, b, e):
        self.s = s
        self.b = b
        self.e = e

    def __or__(self, other):
        assert self.s == other.s
        if isinstance(other, StringSourceTrack):
            return StringNodeSourceTrack(self.s, self.b, other.c+1)
        return StringNodeSourceTrack(self.s, self.b, other.e)

    def __repr__(self):
        return self.s[:self.b]+' . '+self.s[self.b:]
#        return self.s[:self.b]+' [[ '+self.s[self.b:self.e]+' ]] '+self.s[self.e:]
#        return "{} - {}: {}".format(self.b, self.e, self.s)

class StringSourceTrack:
    def __init__(self, s, c):
        self.s = s
        self.c = c

    def __or__(self, other):
        assert self.s == other.s
        if isinstance(other, StringSourceTrack):
            return StringNodeSourceTrack(self.s, self.c, other.c+1)
        return StringNodeSourceTrack(self.s, self.c, other.e)

    def __repr__(self):
        return self.s[:self.c]+' . '+self.s[self.c:]
 #       return "{}: {}".format(self.c, self.s)

class CharLocation:
    def __init__(self, s, b, e):
        self.s = s
        self.b = b
        self.e = e

    def __add__(self, other):
        assert isinstance(other, CharLocation)
        assert self.s == other.s
        assert self.e == other.b
        return CharLocation(self.s, self.b, other.e)

class StringTraceSource (StringTokenSource):
    def get(self):
        if self.eof: return Token('', EOF)
        self.eof = True
        return [Token(c, src=StringSourceTrack(self.s, n))
                for n,c in enumerate(self.s)]
